import asyncio

from pydantic import BaseModel


class State(BaseModel):
    hard_failure: bool = True
    glob: list[str] = []
    include_tar: bool = False
    batch_size: int = 50
    run_id: str | None = None
    pipeline: str | None = None
    build: str | None = None
    runner: asyncio.Runner | None = None

    class Config:
        arbitrary_types_allowed = True



from genomoncology.parse.doctypes import DocType, __TYPE__
from cytoolz.curried import curry
from openpyxl import load_workbook

from .base import Source


def clean_row(row):
    for item in row:
        if isinstance(item, str):
            yield item.strip()
        else:
            yield item


@curry
class ExcelSource(Source):
    def __init__(self, filename):
        self.filename = filename
        self.wb = load_workbook(self.filename)

    def __iter__(self):
        for sheet_name in self.wb.sheetnames:
            ws = self.wb[sheet_name]
            ws_data = ws.values
            columns = next(ws_data)

            for row in ws_data:
                record = dict(zip(columns, clean_row(row)))
                record["sheet_name"] = sheet_name
                record[__TYPE__] = DocType.TSV.value
                yield record



from . import (
    baylor_wgl,
    dbnsfp,
    gnomad,
    hgmd,
    hgmd_gene_list,
    markers,
    clinvar,
    thousandgenomes,
    dbsnp,
    exac,
    splicing,
    baylor_decipher,
    evs,
    cosmic,
    cosmic_non_snv,
    hpo,
    omim,
    genie_maf,
    uniprot_protein_name,
    uniprot_domains,
    uniprot,
    mitomap,
    gene_type,
    gnomad_gene,
    secondary_findings_genes,
    mito_trna,
    mtdb,
    clingen_dosage,
    gnomad_sv,
    gnomad_mnv,
    gnomad_mnv_3bp,
    sv_hash,
)

__all__ = (
    "baylor_wgl",
    "dbnsfp",
    "gnomad",
    "hgmd",
    "hgmd_gene_list",
    "markers",
    "clinvar",
    "thousandgenomes",
    "dbsnp",
    "exac",
    "splicing",
    "evs",
    "cosmic",
    "cosmic_non_snv",
    "baylor_decipher",
    "hpo",
    "omim",
    "baylor_decipher",
    "genie_maf",
    "uniprot_protein_name",
    "uniprot_domains",
    "uniprot",
    "mitomap",
    "gene_type",
    "gnomad_gene",
    "secondary_findings_genes",
    "mito_trna",
    "mtdb",
    "gnomad_sv",
    "gnomad_mnv",
    "gnomad_mnv_3bp",
    "clingen_dosage",
    "sv_hash",
)



from cytoolz.curried import curry, dissoc
from . import Sink
import os

try:
    import ujson as json
except ImportError:
    import json


FILE_PATH = "/Users/ian/work/m2gen/markers/Fusion/"


@curry
class TempAlterationSink(Sink):
    def __init__(self, filename):
        # todo: disregarding the filename
        pass

    def close(self):
        pass

    def write(self, unit):
        file_path = unit.get("file_path")
        rna_id = os.path.basename(file_path).split("_")[0]
        f = open(os.path.join(FILE_PATH, f"{rna_id}_fusion.jsonl"), "a")

        unit = dissoc(unit, "__type__", "file_path")
        f.write(json.dumps(unit))
        f.write("\n")
        f.close()



import logging

try:
    import ujson as json

    default_json_kwargs = dict(escape_forward_slashes=False)
except ImportError:
    import json

    default_json_kwargs = {}

from typing import Optional
from pygments import highlight, lexers, formatters

from cytoolz.curried import curry
from functools import singledispatch


@singledispatch
def to_json_str(unit, **kwargs):
    kwargs = {**default_json_kwargs, **kwargs}
    return json.dumps(unit, **kwargs)


@to_json_str.register(str)  # noqa F811
def _(unit, **_):
    return unit


@to_json_str.register(float)  # noqa F811
@to_json_str.register(int)
def _(unit, **_):
    return str(unit)


@curry
def to_pretty_json_str(unit):
    formatted_json = to_json_str(unit, indent=4)
    colorful_json = highlight(
        formatted_json, lexers.JsonLexer(), formatters.TerminalFormatter()
    )
    return colorful_json


@singledispatch
def from_json_str(unit):
    return unit


@from_json_str.register(str)  # noqa F811
def _(unit):
    return json.loads(unit)


@singledispatch
def clean_str(unit):
    raise NotImplementedError(f"{unit}: Invalid Type [{type(unit)}]")


@clean_str.register(str)  # noqa F811
def _(unit):
    return unit.strip()


def non_null(unit):
    return unit is not None and unit != ""


@singledispatch
def to_str(value):
    return str(value)


@to_str.register(float)  # noqa F811
def _(value):
    return f"{value:.5f}"


def obj_to_dict(fields, obj, **extras) -> Optional[dict]:
    if obj is None:
        return None

    # return all from Bravado object if fields is None
    elif fields is None:
        try:
            return obj._Model__dict
        except Exception as e:
            logging.warning(e)
            return vars()

    # return only selected fields
    else:
        a_dict = {key: getattr(obj, key, None) for key in fields}
        a_dict.update(extras)
        return a_dict


def fint(str_val):
    """ try to convert to int, then float if fails, then return None. """
    for f in [int, float]:
        try:
            return f(str_val)
        except (ValueError, TypeError):
            pass


@curry
def cint(str_val, default=0):
    """ try to convert to int, then float if fails, then return None. """
    try:
        return int(str_val)
    except (ValueError, TypeError):
        return default



from cytoolz.curried import assoc, compose
from genomoncology.parse.doctypes import DocType, __CHILD__, __TYPE__
from genomoncology.pipeline.transformers import register, name_mapping

NAME_MAPPING = {
    "hgvs_g": "hgvs_g",
    "disease__string": "disease",
    "aa_change__string": "aa_change",
    "conservation__string": "conservation",
    "haplogroups__mstring": "haplogroups",
    "loci__string": "loci",
    "fl_count__int": "fl_count",
    "cr_count__int": "cr_count",
    "fl_frequency__float": "fl_frequency",
    "cr_frequency__float": "cr_frequency",
    "variant_haplogroup_counts__mint": "variant_haplogroup_counts",
    "haplogroup_counts__mint": "haplogroup_counts",
    "mitoTIP__float": "mitoTIP",
    "helix_count__int": "helix_count",
    "helix_total_count__int": "helix_total_count",
    "helix_frequency__float": "helix_frequency",
    "gene__string": "gene",
    "apogee__mstring": "apogee",
    "csra": "csra",
}

register(
    input_type=DocType.MITOMAP_RECORD,
    output_type=DocType.MITOMAP,
    transformer=compose(
        lambda x: assoc(x, __TYPE__, DocType.MITOMAP.value),
        name_mapping(NAME_MAPPING),
    ),
)

register(
    input_type=DocType.MITOMAP_RECORD,
    output_type=DocType.MITOMAP,
    transformer=compose(lambda x: assoc(x, __CHILD__, DocType.MITOMAP.value)),
    is_header=True,
)



import asyncio
import logging
import uuid

import click
from click.types import Path
from cytoolz.curried import (concat, curry, filter, map, partition_all,
                             partitionby, unique)
from genomoncology import kms
from genomoncology.cli import options
from genomoncology.cli.state import State
from genomoncology.parse import (DocType, filter_out_ref_unknown,
                                 is_not_skipped_call)
from genomoncology.parse.doctypes import is_annotated_call, is_header
from genomoncology.pipeline import (converters, filters, run_pipeline, sinks,
                                    sources, transformers)
from genomoncology.pipeline.filters import (
    filter_bnd_cpx_ctx_lns_from_load_annotations,
    filter_del_lines_from_load_annotations)

from govcf import BEDFilter, iterate_vcf_files
import gosdk

VERSION = "0.11.3"
COPYRIGHT = "Copyright: GenomOncology, LLC"

_logger = logging.getLogger(__name__)


def print_version(ctx, _, value):
    if not value or ctx.resilient_parsing:
        return
    click.echo(f"GO CLI: Version {VERSION}")
    ctx.exit()


@click.group(chain=True, invoke_without_command=True)
@click.argument("input", type=str)
@click.argument("output", type=str)
@click.option(
    "--dir",
    "dir",
    help="Use this to specify the directory "
    "where your input and output "
    "files are located.",
)
@click.option(
    "--version",
    is_flag=True,
    callback=print_version,
    expose_value=False,
    is_eager=True,
)
@options.common_options
@click.pass_context
def gocli(ctx, input, output, dir):
    logging.basicConfig(level=logging.INFO)
    ctx.ensure_object(State)
    ctx.obj.runner = ctx.with_resource(asyncio.Runner())
    gosdk.setup_sdk()


gocli.result_callback()(run_pipeline)


@gocli.command(name="from_vcf")
@click.option(
    "--bed",
    "-b",
    help="BED file for filtering VCF records.",
    type=Path(exists=True),
)
@click.option(
    "--padding",
    "-p",
    help="Padding added to BED regions.",
    type=int,
    default=0,
)
@click.option(
    "--keep_ref_unknown_calls",
    "-k",
    "keep_ref_unknown_calls",
    help="Include homozygous reference and unknown GT variants as calls",
    is_flag=True,
)
@options.glob_option
@options.build_option
@click.pass_context
def from_vcf(ctx, bed, padding, keep_ref_unknown_calls):
    """Parses VCF into variant calls objects."""
    bed_filter = BEDFilter(bed, padding=padding) if bed else None
    return [
        sources.CollectFilePathsSource(glob=ctx.obj.glob),
        curry(iterate_vcf_files)(
            include_vaf=True,
            bed_filter=bed_filter,
            build=ctx.obj.build,
            keep_ref_unknown_calls=keep_ref_unknown_calls,
        ),
    ]


@gocli.command(name="from_bed")
@click.option(
    "--genes",
    "-g",
    "gene_column",
    help="Specify if bed file has gene column",
    is_flag=True,
)
@click.option("--skip_comment", type=bool, default=False)
@click.option("--comment_char", type=str, default="#")
@options.build_option
@click.pass_context
def from_bed(ctx, gene_column, skip_comment, comment_char):
    """ Read records from a BED file source. """
    columns = ["chromosome", "start", "end"]
    if gene_column:
        columns.append("gene")
    return [
        sources.DelimitedFileSource(
            columns=columns,
            delimiter="\t",
            build=ctx.obj.build,
            skip_comment=skip_comment,
            comment_char=comment_char,
        )
    ]


@gocli.command(name="from_tsv")
@click.option(
    "--column",
    "-c",
    "columns",
    help="Specify columns of TSV source.",
    type=str,
    multiple=True,
)
@click.option("--skip_comment", type=bool, default=False)
@click.option("--comment_char", type=str, default="#")
@click.option("--include_header", type=bool, default=True)
@options.build_option
@click.pass_context
def from_tsv(ctx, columns, skip_comment, comment_char, include_header):
    """Parses TSV into name-value pair objects."""
    return [
        sources.DelimitedFileSource(
            columns=columns,
            delimiter="\t",
            build=ctx.obj.build,
            skip_comment=skip_comment,
            comment_char=comment_char,
            include_header=include_header,
        )
    ]


@gocli.command(name="from_xml")
@click.argument("data_set_name", type=str)
@options.glob_option
@options.build_option
@click.pass_context
def from_xml(ctx, data_set_name):
    return [
        sources.CollectFilePathsSource(glob=ctx.obj.glob),
        sources.XMLSource(data_set_name=data_set_name),
    ]


@gocli.command(name="aggregate_tsv")
@click.argument("aggregate_key", type=str)
@click.option("--backup_key", type=str, default=None)
@options.build_option
@click.pass_context
def aggregate_tsv(ctx, aggregate_key, backup_key):
    """ Rolls up TSV records using a specified key. """
    return [
        sources.AggregatedFileSource(
            aggregate_key=aggregate_key, delimiter="\t", backup_key=backup_key
        )
    ]


@gocli.command(name="aggregate_csv")
@click.argument("aggregate_key", type=str)
@options.build_option
@click.pass_context
def aggregate_csv(ctx, aggregate_key):
    """ Rolls up CSV records using a specified key. """
    return [
        sources.AggregatedFileSource(
            aggregate_key=aggregate_key, delimiter=","
        )
    ]


@gocli.command(name="aggregate_omim_tsv")
@options.build_option
@click.pass_context
def aggregate_omim_tsv(ctx):
    """ Aggregate OMIM TSV file. """
    return [sources.AggregatedOmimFileSource()]


@gocli.command(name="aggregate_cosmic")
@click.argument("cosmic_tsv_path", type=str)
@options.build_option
@click.pass_context
def aggregate_cosmic(ctx, cosmic_tsv_path):
    """ Aggregate COSMIC files. Takes VCF and TSV file. """
    return [sources.AggregatedCOSMICSources(cosmic_tsv=cosmic_tsv_path)]


@gocli.command(name="aggregate_non_snv_cosmic")
@click.argument("record_type", type=click.Choice(["cnv", "fusion"]))
@options.build_option
@click.pass_context
def aggregate_non_snv_cosmic(ctx, record_type):
    return [sources.AggregatedCOSMICNonSNVSources(record_type=record_type)]


@gocli.command(name="from_csv")
@click.option(
    "--column",
    "-c",
    "columns",
    help="Specify columns of TSV source.",
    type=str,
    multiple=True,
)
@options.build_option
@click.pass_context
def from_csv(ctx, columns):
    """Parses CSV into name-value pair objects."""
    return [
        sources.DelimitedFileSource(
            columns=columns, delimiter=",", build=ctx.obj.build
        )
    ]


@gocli.command(name="from_excel")
def from_excel():
    """Parses Excel into name-value pair objects."""
    # todo: merge with parse using --format
    return [sources.ExcelSource]


@gocli.command(name="from_maf")
@click.option(
    "--column",
    "-c",
    "columns",
    help="Specify columns of TSV source.",
    type=str,
    multiple=True,
)
@options.build_option
@click.pass_context
def from_maf(ctx, columns):
    """ Read records from a MAF file source. """
    return [sources.MAFSource(columns=columns, build=ctx.obj.build)]


@gocli.command(name="load_hotspots")
@click.argument("data_set", type=str)
@click.argument("data_set_version", type=str)
@options.size_option
@click.pass_context
def load_hotspots(ctx, data_set="DATASET", data_set_version="DSV"):
    return [
        filter(is_header),
        partition_all(
            ctx.obj.batch_size
        ),  # annotate calls is GET, hardcoded batch size
        kms.async_processor(
            ctx.obj,
            kms.annotations.hotspot_annotate,
            data_set=data_set,
            data_set_version=data_set_version,
        ),
        concat,
    ]


@gocli.command(name="annotate_calls")
@click.option(
    "--fields",
    "-f",
    help="Add field to standard annotation set.",
    type=str,
    multiple=True,
)
@click.option(
    "--delete-if-exists",
    "-d",
    "delete_if_exists",
    help="Delete if annotations exist.",
    is_flag=True,
)
@click.option(
    "--keep_ref_unknown_calls",
    "-k",
    "keep_ref_unknown_calls",
    help="Include homozygous reference and unknown calls.",
    is_flag=True,
)
@click.option(
    "--filter_file",
    "-ff",
    help="Annotations filter file",
    type=str,
    default=None,
)
@click.option(
    "--hgvs_batch",
    "-h",
    "hgvs_batch",
    help="Input is a batch of g. or c. strings instead of CALL or VARIANT "
    "object",
    is_flag=True,
)
@options.size_option
@options.build_option
@click.pass_context
def annotate_calls(
    ctx,
    fields,
    delete_if_exists,
    keep_ref_unknown_calls,
    filter_file,
    hgvs_batch,
):
    """Get annotations for stream of calls."""
    filter_anns = (
        [ann for ann in list(sources.AnnotationsFilterFileSource(filter_file))]
        if filter_file is not None
        else None
    )

    # get the annotation bundle version once instead of once per batch
    annotation_bundle_version = kms.annotations.get_annotation_bundle_version()

    return [
        filter(is_not_skipped_call),
        filter(
            filter_out_ref_unknown(
                keep_ref_unknown_calls=keep_ref_unknown_calls
            )
        ),
        partition_all(
            ctx.obj.batch_size
        ),  # annotate calls is GET, hardcoded batch size
        kms.async_processor(
            ctx.obj,
            kms.annotations.annotate,
            fields=fields,
            delete_if_exists=delete_if_exists,
            build=ctx.obj.build,
            filter_anns=filter_anns,
            annotation_bundle_version=annotation_bundle_version,
            hgvs_batch=hgvs_batch,
        ),
        concat,
    ]


@gocli.command(name="annotate_match")
@click.option(
    "--delete-if-exists",
    "-d",
    "delete_if_exists",
    help="Delete if annotations exist.",
    is_flag=True,
)
@click.option(
    "--keep_ref_unknown_calls",
    "-k",
    "keep_ref_unknown_calls",
    help="Include homozygous reference and unknown calls.",
    is_flag=True,
)
@click.option(
    "--hgvs_batch",
    "-h",
    "hgvs_batch",
    help="Input is a batch of g. or c. strings instead of CALL or VARIANT "
    "object",
    is_flag=True,
)
@options.size_option
@options.build_option
@click.pass_context
def annotate_match(
    ctx,
    delete_if_exists,
    keep_ref_unknown_calls,
    hgvs_batch,
):
    """Get annotations for stream of calls."""

    # get the annotation bundle version once instead of once per batch
    annotation_bundle_version = kms.annotations.get_annotation_bundle_version()

    return [
        filter(is_not_skipped_call),
        filter(
            filter_out_ref_unknown(
                keep_ref_unknown_calls=keep_ref_unknown_calls
            )
        ),
        partition_all(
            ctx.obj.batch_size
        ),  # annotate calls is GET, hardcoded batch size
        kms.async_processor(
            ctx.obj,
            kms.annotations.annotate_match,
            delete_if_exists=delete_if_exists,
            build=ctx.obj.build,
            annotation_bundle_version=annotation_bundle_version,
            hgvs_batch=hgvs_batch,
        ),
        concat,
    ]


@gocli.command(name="annotate_match_bed")
@click.option(
    "--delete-if-exists",
    "-d",
    "delete_if_exists",
    help="Delete if annotations exist.",
    is_flag=True,
)
@options.size_option
@options.build_option
@click.pass_context
def annotate_match_bed(ctx, delete_if_exists):

    """this function hits the annotations/match endpoint. It
        allows the user to send genes
        to the endpoint for annotation."""

    """Get annotations for stream of calls.
    the sdk_function is what makes the call to the
     api endpoint and returns the data"""
    return [
        filter(is_header),
        partition_all(
            ctx.obj.batch_size
        ),  # annotate calls is GET, hardcoded batch size
        kms.async_processor(
            ctx.obj,
            kms.annotations.annotate_genes,
            delete_if_exists=delete_if_exists,
            build=ctx.obj.build,
        ),
        concat,
    ]


@gocli.command(name="get_variant_interpretations")
@click.option(
    "--delete-if-exists",
    "-d",
    "delete_if_exists",
    help="Delete if annotations exist.",
    is_flag=True,
)
@click.argument("template_name", type=str)
@options.build_option
@click.pass_context
def get_variant_interpretations(ctx, delete_if_exists, template_name):
    """Gets the variant interpretations, annotations,
    and protein effects for the passed in variants."""
    return [
        filter(is_not_skipped_call),
        partition_all(
            ctx.obj.batch_size
        ),  # annotate calls is GET, hardcoded batch size
        map(
            kms.variant_interpretations.get_variant_interpretations(
                delete_if_exists=delete_if_exists,
                build=ctx.obj.build,
                template_name=template_name,
            )
        ),
        concat,
    ]


@gocli.command(name="annotate_genes")
@options.build_option
@click.pass_context
def annotate_genes(ctx):
    """Get gene objects by stream of names."""
    return [
        sources.TextFileSource,
        partition_all(ctx.obj.batch_size),
        map(
            kms.genes.boundaries(
                build=ctx.obj.build
            )
        ),
        concat,
    ]


@gocli.command(name="filter_in")
@click.argument("glom_path", type=str)
@click.argument("cmp", type=str)
@click.argument("value", type=str)
def filter_in(glom_path, cmp, value):
    """Filter in objects that match comparison."""
    return filter(filters.glom_include(glom_path, cmp, value))


@gocli.command(name="filter_out")
@click.argument("glom_path", type=str)
@click.argument("cmp", type=str)
@click.argument("value", type=str)
def filter_out(glom_path, cmp, value):
    """Filter out objects that match comparison."""
    return filter(filters.glom_exclude(glom_path, cmp, value))


@gocli.command()
@click.argument("glom_path", type=str)
@click.argument("cmp", type=str)
@click.argument("value", type=str)
def retain(glom_path, cmp, value):
    """Always keep objects that match comparison."""
    return map(filters.mark_retain(glom_path, cmp, value))


@gocli.command(name="to_tsv")
@click.option(
    "--column",
    "-c",
    help="Specify column for TSV display.",
    type=str,
    multiple=True,
)
@click.option("--include-header", "-h", is_flag=True)
def to_tsv(column, include_header):
    """Render objects to TSV file format."""
    return [sinks.TsvFileSink(columns=column, include_header=include_header)]


@gocli.command(name="to_vcf")
@click.option("--header_file_path", "-hfp", type=str)
@click.option(
    "--historical", "-h", help="Write annotations to VCF.", is_flag=True
)
def to_vcf(header_file_path, historical):
    """Render calls to VCF file format."""
    if historical:
        return [sinks.HistoricalVcfFileSink(header_file_path=header_file_path)]
    return [sinks.VcfFileSink(header_file_path=header_file_path)]


@gocli.command(name="to_excel")
def to_excel():
    """Render objects to Excel file format."""
    return [sinks.ExcelSink()]


@gocli.command(name="to_python")
def to_python():
    """Render python objects without JSON transform."""
    return sinks.TextFileSink


@gocli.command(name="to_pretty")
def to_pretty():
    """Render indented, syntax highlighted JSON."""
    return map(converters.to_pretty_json_str)


@gocli.command(name="match_trials")
@click.option(
    "--diseases",
    "-d",
    help="Disease(s) for match eligibility.",
    type=str,
    multiple=True,
)
@click.option("--gender", help="Gender for match eligibility.", type=str)
@click.option(
    "--dob",
    help="Date of birth for match eligibility. Format: YYYY-MM-DD",
    type=str,
)
@click.pass_context
def match_trials(ctx, diseases, gender, dob):
    """Match trials by variants and disease."""
    from . import utils

    utils.check_dob_format(dob)
    return [
        partition_all(99_999_999),
        map(
            kms.trials.match_trials(
                diseases=diseases,
                gender=gender,
                dob=dob,
            )
        ),
        concat,
    ]


@gocli.command(name="match_therapies")
@click.option(
    "--diseases",
    "-d",
    help="Disease(s) for match eligibility.",
    type=str,
    multiple=True,
)
@click.pass_context
def match_therapies(ctx, diseases):
    """Match therapies by variants and disease."""
    return [
        partition_all(99_999_999),
        map(
            kms.therapies.match_therapies(
                diseases=diseases
            )
        ),
        concat,
        filter(converters.non_null),
    ]


@gocli.command(name="sv_match")
@click.option(
    "--delete-if-exists",
    "-d",
    help="Delete if exists, before recalculating annotation.",
    type=bool,
    multiple=False,
    default=False,
)
@click.option(
    "--region_match_tolerance",
    "-rmt",
    help="Region match tolerance",
    default=0.5,
)
@click.option(
    "--cnv_match_tolerance", "-cmt", help="CNV match tolerance", default=0.33
)
@click.option(
    "--keep_ref_unknown_calls",
    "-k",
    "keep_ref_unknown_calls",
    help="Include homozygous reference and unknown calls.",
    is_flag=True,
)
@options.build_option
@click.pass_context
def sv_match(
    ctx,
    delete_if_exists,
    region_match_tolerance,
    cnv_match_tolerance,
    keep_ref_unknown_calls,
):
    """Calls the annotations sv matching endpoint."""
    return [
        filter(
            filter_out_ref_unknown(
                keep_ref_unknown_calls=keep_ref_unknown_calls
            )
        ),
        filter(is_not_skipped_call),
        partition_all(99_999_999),
        map(
            kms.annotations.annotate_sv_match(
                delete_if_exists=delete_if_exists,
                build=ctx.obj.build,
                region_match_tolerance=region_match_tolerance,
                cnv_match_tolerance=cnv_match_tolerance,
            )
        ),
        concat,
    ]


@gocli.command(name="mega_match")
@click.option(
    "--diseases", "-d", help="Diseases", type=str, multiple=True, required=True
)
@click.option(
    "--negative_alterations",
    "-n",
    help="Alterations that were not detected.",
    type=str,
    multiple=True,
)
@click.option(
    "--indeterminate_alterations",
    "-i",
    help="List of alterations that are indeterminate "
    "(i.e. considered not tested).",
    type=str,
    multiple=True,
)
@click.option(
    "--evidence_category_filter",
    "-ev",
    help='Set to "exclude_preclinical" to exclude preclinical evidence',
    type=str,
)
@click.option(
    "--exclude_non1A_negative_matches",
    "-ex",
    help="Exclude wild-type non-1A negative therapy matches",
    is_flag=True,
)
@click.option(
    "--remove_nbm_trials_if_no_drug_target",
    "-r",
    help="Remove all non-biomarker trials from therapies that "
    "have no gene targets or associations",
    is_flag=True,
)
@click.option(
    "--therapeutic_context_filter",
    "-th",
    help='Set to "exclude_adjuvant" to filter out therapies and trials '
    "that have a therapeutic context of adjuvant or neoadjuvant",
    type=str,
)
@click.option(
    "--therapy_type_filter",
    "-tt",
    help='Set to "exclude_non_targeted_therapies" to filter out therapies '
    "and trials that don’t have a highest level drug type of targeted "
    "therapy or immunotherapy",
    type=str,
)
@click.option(
    "--gender",
    "-g",
    help="Filter results on gender. Possible values are Male or Female",
    type=str,
)
@click.option(
    "--date_of_birth",
    "-dob",
    help="Date of birth for match eligibility. Format: YYYY-MM-DD",
    type=str,
)
@click.option(
    "--countries",
    "-c",
    help="Filter results on countries",
    type=str,
    multiple=True,
)
@click.option(
    "--snv_panel", "-snv", help="Genes on SNV panel.", type=str, multiple=True
)
@click.option(
    "--cnv_panel", "-cnv", help="Genes on CNV panel.", type=str, multiple=True
)
@click.option(
    "--cnv_panel_alterations",
    "-cnv_a",
    help="Alterations on CNV panel.",
    type=str,
    multiple=True,
)
@click.option(
    "--ctx_panel",
    "-ctx",
    help="Alterations on CTX panel.",
    type=str,
    multiple=True,
)
@click.option(
    "--karyotype_panel",
    "-kary",
    help="Alterations on Karyotype panel.",
    type=str,
    multiple=True,
)
@click.option(
    "--biomarker_panel",
    "-bio",
    help="Alterations on Biomarker panel.",
    type=str,
    multiple=True,
)
@click.option(
    "--prx_panel", "-prx", help="Genes on PRX panel.", type=str, multiple=True
)
@click.option(
    "--assay", "-a", help="Name(s) of assays", type=str, multiple=True
)
@click.pass_context
def mega_match(ctx, **kwargs):
    """Call mega-match using VIE response and diseases."""
    return [
        filter(is_annotated_call),
        partition_all(99_999_999),
        map(
            kms.mega_match.mega_match(
                **kwargs
            )
        ),
    ]


@gocli.command(name="match_contents")
@click.option(
    "--diseases",
    "-d",
    help="Disease(s) for match eligibility.",
    type=str,
    multiple=True,
)
@click.pass_context
def match_contents(ctx, diseases):
    """Match contents by variants and disease."""
    return [
        partition_all(99_999_999),
        map(
            kms.contents.match_contents(
                diseases=diseases
            )
        ),
        concat,
        filter(converters.non_null),
    ]


@gocli.command()
@click.argument("output_type", type=str)
@options.pipeline_option
@options.run_id_option
@click.pass_context
def transform(ctx, output_type):
    """Transform input stream to output type."""
    return [
        filter(is_not_skipped_call),
        transformers.create_transformer(output_type=output_type, state=ctx.obj),
        filter(None),
    ]


@gocli.command()
@click.argument("field", type=str)
def extract(field):
    """Extract a specific field from object stream."""
    return map(transformers.get_in_field(field))


@gocli.command()
def flatten():
    """Flatten list of lists into a list of strings."""
    return [filter(converters.non_null), concat]


@gocli.command()
def distinct():
    """Remove duplicates from a list of strings."""
    return unique


@gocli.command(name="load_warehouse")
@click.option(
    "--workers",
    help="Number of load workers (default: 10)",
    type=int,
    default=10,
)
@click.option(
    "--keep_ref_unknown_calls",
    "-k",
    "keep_ref_unknown_calls",
    help="Include homozygous reference and unknown calls.",
    is_flag=True,
)
@options.size_option
@options.build_option
@click.pass_context
def load_warehouse(ctx, workers, keep_ref_unknown_calls):
    """Loads variants to warehouse."""
    return [
        filter(is_not_skipped_call),
        filter(
            filter_out_ref_unknown(
                keep_ref_unknown_calls=keep_ref_unknown_calls
            )
        ),
        partition_all(ctx.obj.batch_size),
        sinks.LoadWarehouseVariantsSink(state=ctx.obj, num_workers=workers),
        concat,
    ]


@gocli.command(name="load_features")
@click.option(
    "--workers",
    help="Number of load workers (default: 10)",
    type=int,
    default=10,
)
@options.size_option
@options.build_option
@click.pass_context
def load_features(ctx, workers):
    """Loads variants to warehouse."""
    return [
        partition_all(ctx.obj.batch_size),
        sinks.LoadWarehouseFeaturesSink(state=ctx.obj, num_workers=workers),
        concat,
    ]


@gocli.command(name="load_custom")
@click.argument("data_set", type=str)
@click.argument("data_set_version", type=str)
@click.option(
    "--workers",
    help="Number of load workers (default: 10)",
    type=int,
    default=10,
)
@click.pass_context
def load_custom(ctx, data_set, data_set_version, workers):
    """This command hits the annotations/load_custom endpoint.
        This leverages the LoadAnnotationsSink and sends through
        a is_custom boolean. When true, it hits the load_custom endpoint,
        when False, it hits the load_annotations endpoint."""
    filter_step = filter_del_lines_from_load_annotations
    return [
        filter(is_not_skipped_call),
        filter(filter_step),
        transformers.create_transformer(output_type=data_set, state=ctx.obj),
        filter(None),
        partition_all(ctx.obj.batch_size),
        sinks.LoadAnnotationSink(
            state=ctx.obj,
            data_set=data_set,
            data_set_version=data_set_version,
            num_workers=workers,
            is_custom=True
        ),
        concat,
        sinks.JsonlFileSink,
    ]


@gocli.command(name="load_annotations")
@click.argument("data_set", type=str)
@click.argument("data_set_version", type=str)
@click.option(
    "--workers",
    help="Number of load workers (default: 10)",
    type=int,
    default=10,
)
@options.size_option
@options.build_option
@click.pass_context
def load_annotations(ctx, data_set, data_set_version, workers):
    """Loads variant objects into annotations core."""

    """ In a previous ticket there was logic that was implemented to remove
     any lines in the vcf file that have <DEL> the reasoning was due to the
     fact that we do not yet have the ability to process them. As of kms-4333
     the gnomad_sv dataset will require the ability to process these types and
    therefore we created 2 options for the building of the pipeline.
    If the dataset is gnomad_sv then build the pipeline without filtering out
    the <DEL>'s otherwise filter out the <DEL>."""
    if data_set == "gnomad_sv":
        filter_step = filter_bnd_cpx_ctx_lns_from_load_annotations
    else:
        filter_step = filter_del_lines_from_load_annotations
    return [
        filter(is_not_skipped_call),
        filter(filter_step),
        transformers.create_transformer(output_type=data_set, state=ctx.obj),
        filter(None),
        partition_all(ctx.obj.batch_size),
        sinks.LoadAnnotationSink(
            state=ctx.obj,
            data_set=data_set,
            data_set_version=data_set_version,
            num_workers=workers,
        ),
        concat,
        sinks.JsonlFileSink,
    ]


@gocli.command(name="refresh_annotations")
@click.option(
    "--hgvs", "recalculate_hgvs", help="Full HGVS refresh.", is_flag=True
)
@click.option("--bed", "-b", help="Bed file", type=str, default=None)
@options.size_option
@click.pass_context
def refresh_annotations(ctx, recalculate_hgvs, bed):
    """Rebuilds annotations merged core."""
    source = sources.BedFileSource(bed_file=bed) if bed else sources.NullSource
    return [
        source,
        map(
            kms.annotations.refresh(
                recalculate_hgvs=recalculate_hgvs,
                batch_size=ctx.obj.batch_size,
            )
        ),
    ]


@gocli.command(name="region_search")
@click.option("--genes", type=bool, default=False)
@click.pass_context
def region_search(ctx, genes):
    """Searches for Transcript in region specified
    by bed file"""
    """if genes is True then we will
            return the names of the genes
          that are processed in the bed file. These
          can then be used for further processing
          for instance using the annotate_match_bed command
           which will annotate them."""

    process_func = (
        kms.transcripts.process_transcript_batch_get_genes
        if genes
        else kms.transcripts.process_transcript_batch
    )

    return [
        filter(DocType.HEADER.is_not),
        partition_all(ctx.obj.batch_size),
        map(process_func),
        concat,
    ]


@gocli.command(name="from_source")
@click.option(
    "--arg", "-a", "args", help="Argument: name=value", type=str, multiple=True
)
@click.argument("source_name", type=str)
@click.pass_context
def from_source(ctx, source_name, args):
    """Parse file into name-value pair objects."""
    # TODO: good enough for uniprot and mitomap, but refactor later
    source = sources.get_one_off_source(source_name)
    assert source, f"Source not found: {source_name}"
    kwargs = {}
    if args:
        kwargs = dict([a.split("=") for a in args])
    return [source(build=ctx.obj.build, **kwargs)]


@gocli.command()
@click.option(
    "--arg", "-a", "args", help="Argument: name=value", type=str, multiple=True
)
@click.argument("function_name", type=str)
def invoke(function_name, args):
    """Invoke an external function."""
    import importlib

    module_name, function_name = function_name.rsplit(".", 1)
    module = importlib.import_module(module_name)

    f = getattr(module, function_name)
    f = curry(f)

    if args:
        kwargs = dict([a.split("=") for a in args])
        f = f(**kwargs)

    return map(f)


@gocli.command(name="add_flag")
@click.argument("bed_file", type=str)
@click.argument("flag_name", type=str)
def add_flag(bed_file, flag_name):
    """Add flag based on BED file value/presence."""
    bed_filter = BEDFilter(bed_file)
    return map(transformers.add_flag(bed_filter, flag_name))


@gocli.command()
def version():
    print(f"GO CLI: Version {VERSION} [{COPYRIGHT}]")
    return [sources.NullSource]


@gocli.command()
def rollup():
    """Roll up variants by record."""
    return [
        partitionby(lambda r: r.get("__record__", uuid.uuid4())),
        map(transformers.add_rollup),
        concat,
    ]



from cytoolz.curried import curry, assoc

from genomoncology.parse import ensures, DocType, __TYPE__, __CHILD__
from .base import LazyFileSource


@curry
class MAFSource(LazyFileSource):
    def __init__(self, filename, columns, **meta):
        self._mapper = None
        self.columns = ensures.ensure_collection(columns)
        self.meta = meta
        self.delimiter = "\t"

        if __TYPE__ not in meta:
            self.meta = assoc(self.meta, __TYPE__, DocType.MAF.value)

        super().__init__(filename)

    def __iter__(self):
        # noinspection PyUnresolvedReferences
        iterator = super(MAFSource.func, self).__iter__()

        while not self.columns:
            line = next(iterator)
            if line[0] == "#":
                continue
            self.columns = line.strip().split("\t")

        yield self.create_header()

        for record in iterator:
            record = record.replace("\n", "").split("\t")
            ref_index = self.columns.index("Reference_Allele")
            alt1_index = self.columns.index("Tumor_Seq_Allele1")
            alt2_index = self.columns.index("Tumor_Seq_Allele2")
            for alt_index in [alt1_index, alt2_index]:
                if record[ref_index] != record[alt_index]:
                    yield self.create_maf_call(
                        record, alt_index, self.meta["build"]
                    )

    def create_header(self):
        return {
            __TYPE__: DocType.HEADER.value,
            __CHILD__: self.meta.get(__TYPE__),
            "columns": self.columns,
            "meta": self.meta,
            "file_path": self.name,
        }

    def create_maf_call(self, record, alt_index, build):
        maf_d = dict(__type__=DocType.MAF.value, alt=record[alt_index])
        for column in self.columns:
            maf_d[column] = record[self.columns.index(column)]
        maf_d["Chromosome"] = maf_d["Chromosome"].replace("chr", "")
        maf_d["build"] = build
        return maf_d



from .registry import create_transformer, register
from .functions import (
    filter_keys,
    filter_private_keys,
    get_in_field,
    name_mapping,
    rename,
    transform,
    is_multi_field,
    split_value,
    add_flag,
    add_rollup,
    get_boolean_value,
)
from .mapper_classes import VariantMapper

from . import tx

__all__ = (
    "VariantMapper",
    "create_transformer",
    "filter_keys",
    "filter_private_keys",
    "get_in_field",
    "name_mapping",
    "register",
    "rename",
    "tx",
    "transform",
    "is_multi_field",
    "split_value",
    "add_flag",
    "add_rollup",
    "get_boolean_value",
)



from cytoolz.curried import assoc, compose
from genomoncology.parse.doctypes import DocType, __CHILD__, __TYPE__
from genomoncology.pipeline.transformers import register, name_mapping

NAME_MAPPING = {
    # hgvs
    "chr": "Chromosome",
    "start": "Start_Position",
    "ref": "Reference_Allele",
    "alt": "alt",
    # other data
    "sample_name": "Tumor_Sample_Barcode",
    "ref_depth": "t_ref_count",
    "alt_depth": "t_alt_count",
    "build": "build",
}

register(
    input_type=DocType.MAF,
    output_type=DocType.VARIANT,
    transformer=compose(
        lambda x: assoc(x, "file_path", "GENIE"),
        lambda x: assoc(x, "run_id", "GENIE"),
        lambda x: assoc(x, "start", int(x["start"])),
        lambda x: assoc(x, __TYPE__, DocType.VARIANT.value),
        name_mapping(NAME_MAPPING),
    ),
)

register(
    input_type=DocType.MAF,
    output_type=DocType.VARIANT,
    transformer=compose(lambda x: assoc(x, __CHILD__, DocType.VARIANT.value)),
    is_header=True,
)



import types


def ensure_collection(x):
    """ Ensures the parameter provided is an iterable """

    if x is None:
        x = []

    elif isinstance(x, types.GeneratorType):
        x = list(x)

    elif not isinstance(x, (list, tuple, set)):
        x = [x]

    return x


# reference: http://stackoverflow.com/a/12472564
def flatten(x):

    if isinstance(x, (tuple, types.GeneratorType)):
        x = list(x)

    if x == []:
        return x

    if isinstance(x[0], (tuple, list, types.GeneratorType)):
        return flatten(x[0]) + flatten(x[1:])

    return x[:1] + flatten(x[1:])



from cytoolz.curried import do, pipe, map
from .sinks import Sink, JsonlFileSink
from .sources import Source, JsonlFileSource
from genomoncology.parse.ensures import flatten
import glob
import os


def do_file_run_pipeline(processors, input, output):
    funcs, sinks = prep_funcs_sinks(processors, output)

    for _ in pipe(input, *funcs):
        pass

    for sink in sinks:
        sink.close()


def run_pipeline(processors, input, output, dir):
    if dir is not None:
        if dir == ".":
            dir = os.getcwd()
        input_file_pattern = os.path.join(dir, input)
        matching_files = glob.glob(input_file_pattern)
        for input_file in matching_files:
            head, tail = os.path.split(input_file)
            if output == "-":
                output_file = "-"
            else:
                output_file = os.path.join(
                    dir, output.replace("*", tail.split(".")[0])
                )
            if output_file == "-" or not os.path.isfile(output_file):
                do_file_run_pipeline(processors, input_file, output_file)
    else:
        do_file_run_pipeline(processors, input, output)


def prep_funcs_sinks(processors, output):
    processors = flatten(processors)

    if not processors or not Source.is_source_class(processors[0]):
        processors.insert(0, JsonlFileSource)

    sinks = []
    funcs = []

    def add_sink(cls):
        sink = cls(output)
        sinks.append(sink)
        funcs.append(map(do(sink.write)))

    for processor in processors:
        if Sink.is_sink_class(processor):
            add_sink(processor)

        else:
            funcs.append(processor)

    if not sinks:
        add_sink(JsonlFileSink)

    return funcs, sinks



import xml.etree.cElementTree as etree
import json

from cytoolz.curried import curry, assoc

from genomoncology.parse import DocType, __TYPE__, __CHILD__
from ..base import LazyFileSource


@curry
class UniprotFileSource(LazyFileSource):
    def __init__(self, filename, **meta):
        self.filename = filename
        self.meta = meta

        if __TYPE__ not in meta:
            self.meta = assoc(
                self.meta, __TYPE__, DocType.UNIPROT_RECORD.value
            )

    def get_features_to_ignore(self):
        return [
            "splice variant",
            "mutagenesis site",
            "sequence variant",
            "non-consecutive residues",
            "non-standard amino acid",
            "non-terminal residue",
            "sequence conflict",
            "strand",
            "unsure residue",
        ]

    def __iter__(self):  # pragma: no mccabe
        yield self.create_header()

        context = etree.iterparse(self.filename, events=["start", "end"])
        # uniprot_id (accessions, but only grab first one)
        uniprot_id = None
        # protein name
        in_protein_rec_name = False
        is_in_component = False
        protein = None
        # gene name
        in_gene_name = False
        gene = None
        # features
        features_to_ignore = self.get_features_to_ignore()
        current_feature = None
        all_features = []
        # protein length
        protein_length = None
        # alternative names
        is_in_alternative_name = True
        alternative_names = []
        # canonical protein sequence ids
        current_uniprot_np_id = None
        no_molecule_canonical_nm_id = []
        no_molecule_canonical_np_id = []
        uniprot_canonical_nm_id = []
        uniprot_canonical_np_id = []
        is_in_refseq = False
        is_canonical_refseq = False
        molecule_tag_present = False
        for event, elem in context:
            tag = elem.tag.replace("{http://uniprot.org/uniprot}", "")
            # accession
            if event == "start" and tag == "accession" and uniprot_id is None:
                uniprot_id = elem.text
                continue
            # protein name
            if event == "start" and tag == "component":
                is_in_component = True
                continue
            if event == "end" and tag == "component":
                is_in_component = False
                continue
            if event == "start" and tag == "recommendedName":
                in_protein_rec_name = True
                continue
            if (
                in_protein_rec_name
                and not is_in_component
                and event == "end"
                and tag == "fullName"
            ):
                protein = elem.text
                in_protein_rec_name = False
                continue
            # alternative name
            if event == "start" and tag == "alternativeName":
                is_in_alternative_name = True
                continue
            if is_in_alternative_name and event == "end" and tag == "fullName":
                alternative_names.append(elem.text)
                is_in_alternative_name = False
                continue
            # gene name
            if event == "start" and tag == "gene":
                in_gene_name = True
                continue
            if (
                in_gene_name
                and event == "end"
                and tag == "name"
                and "type" in elem.attrib
                and elem.attrib["type"] == "primary"
            ):
                gene = elem.text
                in_gene_name = False
                continue
            # features
            if (
                event == "start"
                and tag == "feature"
                and "type" in elem.attrib
                and elem.attrib["type"] not in features_to_ignore
            ):
                current_feature = {"type": elem.attrib["type"]}
                if "description" in elem.attrib:
                    current_feature["description"] = elem.attrib["description"]
                continue
            if (
                current_feature is not None
                and event == "start"
                and tag == "position"
            ):
                current_feature["start"] = int(elem.attrib["position"])
                current_feature["end"] = int(elem.attrib["position"])
                continue
            if (
                current_feature is not None
                and event == "start"
                and tag == "begin"
            ):
                if "position" in elem.attrib:
                    current_feature["start"] = int(elem.attrib["position"])
                else:
                    # we will replace this later
                    current_feature["start"] = None
                continue
            if (
                current_feature is not None
                and event == "start"
                and tag == "end"
            ):
                if "position" in elem.attrib:
                    current_feature["end"] = int(elem.attrib["position"])
                else:
                    # we will replace this later
                    current_feature["end"] = None
                continue
            if (
                event == "end"
                and tag == "feature"
                and "type" in elem.attrib
                and elem.attrib["type"] not in features_to_ignore
            ):
                # add the current feature to the features list
                all_features.append(current_feature)
                current_feature = None
                continue
            # protein length
            if (
                event == "start"
                and tag == "sequence"
                and "length" in elem.attrib
            ):
                protein_length = elem.attrib["length"]
                continue
            # canonical np/nm ids
            if (
                event == "start"
                and tag == "dbReference"
                and elem.attrib["type"] == "RefSeq"
            ):
                is_in_refseq = True
                # get the current uniprot np id (in case it's canonical)
                if current_uniprot_np_id is None:
                    current_uniprot_np_id = elem.attrib["id"]
                continue
            if event == "start" and tag == "molecule" and is_in_refseq:
                molecule_tag_present = True
                # see if this is actually canonical np or not
                the_id = elem.attrib["id"]
                if the_id.endswith("-1"):
                    is_canonical_refseq = True
                    uniprot_canonical_np_id.append(current_uniprot_np_id)
                else:
                    # clear out the current nm that we previously set
                    current_uniprot_np_id = None
                continue
            if (
                event == "start"
                and tag == "property"
                and is_in_refseq
                and (is_canonical_refseq or not molecule_tag_present)
            ):
                # if no molecule tag was present, then also
                # set the canonical np
                if not molecule_tag_present:
                    no_molecule_canonical_np_id.append(current_uniprot_np_id)
                    no_molecule_canonical_nm_id.append(elem.attrib["value"])
                    current_uniprot_np_id = None
                else:
                    # since this is canonical, set the canonical nm id
                    uniprot_canonical_nm_id.append(elem.attrib["value"])
                continue
            if event == "end" and tag == "dbReference" and is_in_refseq:
                is_in_refseq = False
                is_canonical_refseq = False
                molecule_tag_present = False
                current_uniprot_np_id = None
                continue
            # end of entry record
            if event == "end" and tag == "entry":
                # process features (replace None start/ends)
                self.process_features(all_features)

                return_d = {
                    "uniprot_id": uniprot_id,
                    "gene": gene,
                    "protein_full_name": protein,
                    "protein_length": protein_length,
                    "protein_alternate_names": alternative_names,
                    "uniprot_canonical_nm_id": uniprot_canonical_nm_id
                    if len(uniprot_canonical_nm_id) > 0
                    else no_molecule_canonical_nm_id,
                    "uniprot_canonical_np_id": uniprot_canonical_np_id
                    if len(uniprot_canonical_np_id) > 0
                    else no_molecule_canonical_np_id,
                    "features": [
                        json.dumps(feature) for feature in all_features
                    ],
                    "__type__": DocType.UNIPROT_RECORD,
                }
                uniprot_id = None
                gene = None
                protein = None
                is_in_component = False
                in_protein_rec_name = False
                current_feature = None
                all_features = []
                protein_length = None
                alternative_names = []
                uniprot_canonical_nm_id = []
                uniprot_canonical_np_id = []
                no_molecule_canonical_nm_id = []
                no_molecule_canonical_np_id = []
                current_uniprot_np_id = None
                is_canonical_refseq = False
                is_in_refseq = False
                molecule_tag_present = False
                elem.clear()
                yield return_d

    def process_features(self, features):
        # if start/end is none, set it to be whatever the other value is
        # so they are the same. If both are none, no changes
        for feature in features:
            if feature.get("start") is None and feature.get("end") is not None:
                feature["start"] = feature["end"]
            elif (
                feature.get("end") is None and feature.get("start") is not None
            ):
                feature["end"] = feature["start"]

    def create_header(self):
        return {
            __TYPE__: DocType.HEADER.value,
            __CHILD__: self.meta.get(__TYPE__),
            "meta": self.meta,
            "fields": [
                "gene",
                "protein_full_name",
                "uniprot_id",
                "protein_length",
                "alternative_names",
                "uniprot_canonical_nm_id",
                "uniprot_canonical_np_id",
                "features",
            ],
        }



from .factory import async_processor
from . import (
    annotations,
    contents,
    genes,
    mega_match,
    therapies,
    transcripts,
    trials,
    variant_interpretations,
)

__all__ = (
    "async_processor",
    "annotations",
    "contents",
    "genes",
    "mega_match",
    "therapies",
    "transcripts",
    "trials",
    "variant_interpretations",
)



from cytoolz.curried import assoc, compose, remove, pipe, merge_with, merge
from genomoncology.parse.doctypes import DocType, __CHILD__, __TYPE__
from genomoncology.pipeline.transformers import register, name_mapping
import re


NAME_MAPPING = {
    "gene": "gene",
    "gene_mim": "MIM Number",
    "phenotype__mstring": "phenotype",
    "phenotype_mim__mstring": "phenotype_mim",
    "inheritance__mstring": "inheritance",
}


def split_filter(value, delimiter, pred_func):
    return filter(pred_func, value.split(delimiter))


def split_remove(value, delimiter, pred_func):
    return remove(pred_func, value.split(delimiter))


def parse_omim_inheritance(omim_pheno_d):
    omim_inheritance_l = [
        "Multifactorial",
        "autosomal dominant",
        "X-linked dominant",
        "X-linked",
        "X-linked recessive",
        "Y-linked",
        "autosomal recessive",
        "Autosomal dominant",
        "Autosomal recessive",
        "Mitochondrial",
        "?Autosomal dominant",
        "Isolated cases",
        "Somatic mutation",
        "Digenic recessive",
        "Somatic mosaicism",
        "?X - linked recessive",
        "Inherited chromosomal imbalance",
        "Pseudoautosomal dominant",
        "Pseudoautosomal recessive",
        "Digenic dominant",
    ]

    # structure of the phenotype field is:
    # {phenotype strings and mim string} {(digit)} {inheritances}
    # so split on (digit) and take the stuff that comes after
    phen = omim_pheno_d["phenotype"]
    omim_inheritance_strings = re.compile(r"\(\d\)").split(phen)[-1]
    phen_before_inheritance = phen[
        0 : len(phen) - len(omim_inheritance_strings)
    ]
    omim_inheritance_strings = omim_inheritance_strings.strip()
    omim_inheritance_values = split_filter(
        omim_inheritance_strings, ", ", lambda x: x in omim_inheritance_l
    )
    omim_pheno_d = assoc(
        omim_pheno_d, "inheritance", ";".join(omim_inheritance_values)
    )

    omim_leftovers = split_remove(
        omim_inheritance_strings, ", ", lambda x: x in omim_inheritance_l
    )

    phen = phen_before_inheritance + ", ".join(omim_leftovers)
    omim_pheno_d["phenotype"] = phen

    return omim_pheno_d


def parse_omim_position(omim_pheno_d):
    omim_position_l = ["(1)", "(2)", "(3)", "(4)"]
    omim_pheno_d = assoc(
        omim_pheno_d,
        "phenotype",
        " ".join(
            split_remove(
                omim_pheno_d["phenotype"], " ", lambda x: x in omim_position_l
            )
        ),
    )
    return omim_pheno_d


def parse_omim_pheno_id(omim_pheno_d):
    if omim_pheno_d["phenotype"][-6:].isdigit():
        omim_pheno_d = assoc(
            omim_pheno_d, "phenotype_mim", omim_pheno_d["phenotype"][-6:]
        )
        omim_pheno_d = assoc(
            omim_pheno_d, "phenotype", omim_pheno_d["phenotype"][:-8]
        )
    else:
        omim_pheno_d = assoc(omim_pheno_d, "phenotype_mim", "")
    return omim_pheno_d


def phenotype_pipeline(x):
    return pipe(
        {"phenotype": x},
        parse_omim_inheritance,
        parse_omim_position,
        parse_omim_pheno_id,
    )


def parse_omim_phenotypes(x):
    if "Phenotypes" in x:

        if all([phen == "None" for phen in x["Phenotypes"]]):
            # do not add any of the phenotypes fields if they are all None
            return x

        # filter out Nones
        x["Phenotypes"] = [phen for phen in x["Phenotypes"] if phen != "None"]

        x["Phenotypes"] = "; ".join(x["Phenotypes"])
        return merge_with(
            list, map(phenotype_pipeline, x["Phenotypes"].split("; "))
        )
    else:
        return x


def parse_omim_gene(x):
    return x["key"]


def replace_empty_or_null_strings_with_none(omim_ann):
    phenotype_values = []
    keys_that_need_values_changed = ("phenotype__mstring",
                                     "phenotype_mim__mstring",
                                     "inheritance__mstring")
    for key in filter(lambda x: x in keys_that_need_values_changed,
                      omim_ann.keys()):
        # first add the values to phenotype_values because if they are None,
        # they will get converted to "None" in the next step.
        phenotype_values.extend(omim_ann[key])
        omim_ann[key] = ["None" if not v else v for v in omim_ann[key]]
    # if all phenotyp_values are None then we want to remove the 3
    # __mstring fields in other words, if all values for all keys that end in
    # __mstring are None then
    # remove all of the keys. However, if one or more of the keys do not have
    # all None values then do not remove.
    if not any(phenotype_values):
        omim_ann.pop('phenotype__mstring')
        omim_ann.pop('phenotype_mim__mstring')
        omim_ann.pop('inheritance__mstring')
    return omim_ann


register(
    input_type=DocType.TSV,
    output_type=DocType.OMIM,
    transformer=compose(
        lambda x: replace_empty_or_null_strings_with_none(x),
        lambda x: assoc(x, "is_gene_annotation", True),
        lambda x: assoc(x, __TYPE__, DocType.OMIM.value),
        name_mapping(NAME_MAPPING),
        lambda x: merge(x, parse_omim_phenotypes(x)),
        lambda x: assoc(x, "gene", parse_omim_gene(x)),
        # name_mapping(NAME_MAPPING_PRE),
    ),
)

register(
    input_type=DocType.TSV,
    output_type=DocType.OMIM,
    transformer=compose(
        lambda x: assoc(x, "is_gene_annotation", True),
        lambda x: assoc(x, __CHILD__, DocType.OMIM.value),
        # lambda x: merge(x, parse_omim_phenotypes(x)),
    ),
    is_header=True,
)



from cytoolz.curried import curry

from genomoncology.pipeline.sources.base import LazyFileSource


@curry
class AnnotationsFilterFileSource(LazyFileSource):
    def __init__(self, ann_file):
        assert ann_file, "Annotations filter file not provided"
        self.file = open(ann_file, "r")
        self.entries = self.file.readlines()

    def __iter__(self):
        for annotation_field in self.entries:
            yield annotation_field.strip()
        self.file.close()



from functools import partial

from cytoolz.curried import curry
from genomoncology.cli.const import GRCH37
from genomoncology.parse.doctypes import DocType
from genomoncology.parse.ensures import ensure_collection
from genomoncology.pipeline.converters import obj_to_dict

import gosdk


@curry
def genes(data, fields=("chromosome", "start", "end", "name")):
    batch = ensure_collection(data)
    gene_list = gosdk.call_with_retry(
        gosdk.sdk.genes.get_genes,
        name=batch,
        fields=fields,
        page_size=len(batch)
    )
    func = partial(obj_to_dict, fields, __type__=DocType.GENE.value)
    return list(map(func, gene_list))


@curry
def boundaries(
    data, fields=("chromosome", "start", "end", "gene"), build=GRCH37
):
    batch = ensure_collection(data)
    gene_list = gosdk.call_with_retry(
        gosdk.sdk.genes.gene_boundaries,
        name=batch,
        build=build
    )
    func = partial(obj_to_dict, fields, __type__=DocType.GENE.value)
    formatted_results = list(map(func, gene_list))
    if gene_list.not_found:
        formatted_results.append("The following genes could not be found: {}".format(
            ", ".join(gene_list.not_found)))
    return formatted_results



import logging

from cytoolz.curried import curry

import gosdk

_logger = logging.getLogger(__name__)


@curry
def process_transcript_batch(data):
    batch = [
        "%s|%s|%s" % (record["chromosome"], record["start"], record["end"])
        for record in data
    ]
    _logger.debug("get_transcripts_batch: batch=%s", batch)

    results = gosdk.call_with_retry(
        gosdk.sdk.region_search.region_search_batch,
        batch=batch
    )
    return results


@curry
def process_transcript_batch_get_genes(data):
    batch = [
        "%s|%s|%s" % (record["chromosome"], record["start"], record["end"])
        for record in data
    ]
    _logger.debug("get_transcripts_batch: batch=%s", batch)
    # Get all regions for the genes
    results = gosdk.call_with_retry(
        gosdk.sdk.region_search.region_search_batch,
        batch=batch
    )
    # we only need the gene names from the result that is returned,
    # We also do not want duplicate names.
    list_of_genes = []
    for line in results:
        transcripts = line.get("transcripts", [])
        for transcript in transcripts:
            list_of_genes.append(transcript.get("gene"))
    return set(list_of_genes)



from cytoolz.curried import curry

from genomoncology.kms.match import vie_results_to_alts
import gosdk

PARAM_NAME_CHANGES = {
    "exclude_non1a_negative_matches": "exclude_non1A_negative_matches"
}


@curry
def mega_match(vies, **kwargs):
    # get the alterations
    alterations, benign_alterations = vie_results_to_alts(vies)

    # convert any parameter names that were changed by click
    for param_name, updated_name in PARAM_NAME_CHANGES.items():
        if param_name in kwargs:
            kwargs[updated_name] = kwargs.pop(param_name, None)

    mega_match_response = gosdk.call_with_retry(
        gosdk.sdk.therapies.mega_match_post,
        alterations=alterations,
        benign_alterations=benign_alterations,
        **kwargs,
    )

    return convert_mega_match_response(mega_match_response)


def convert_mega_match_response(mm_response):
    if mm_response:
        mm_response = mm_response.convert_to_dict()
    return mm_response



from cytoolz.curried import curry, assoc

from genomoncology.parse import DocType
from . import comparators


RETAIN = "__RETAIN__"


def is_keep(obj):
    return obj.get(RETAIN, False) or DocType.HEADER.is_a(obj)


@curry
def or_(filter_1, filter_2, obj):
    return filter_1(obj) or filter_2(obj)


@curry
def or_not2_(filter_1, filter_2, obj):
    return filter_1(obj) or not filter_2(obj)


def glom_include(glom_path: str, cmp_str: str, cmp_value: str):
    glom_comparator = get_glom_comparator(glom_path, cmp_str, cmp_value)
    return or_(is_keep, glom_comparator)


def glom_exclude(glom_path: str, cmp_str: str, cmp_value: str):
    glom_comparator = get_glom_comparator(glom_path, cmp_str, cmp_value)
    return or_not2_(is_keep, glom_comparator)


def get_glom_comparator(glom_path: str, cmp_str: str, cmp_value: str):
    cmp_str = cmp_str.upper()
    cmp_set = comparators.COMPARE_LOOKUP.get(cmp_str, None)

    if cmp_set:
        return comparators.compare(glom_path, cmp_set, cmp_value)

    if cmp_str in "IN":
        cmp_value = load_data_set(cmp_value)
        return comparators.compare_in(glom_path, cmp_value)

    if cmp_str in ("HAS", "CONTAINS"):
        return comparators.compare_contains(glom_path, cmp_value)


def load_data_set(literal_str_or_file_path: str) -> set:
    if literal_str_or_file_path.startswith("@"):
        file_path = literal_str_or_file_path[1:]
        return set([line.strip() for line in open(file_path).readlines()])
    else:
        return set(
            [string.strip() for string in literal_str_or_file_path.split(",")]
        )


@curry
def mark_retain(glom_path: str, cmp_str: str, cmp_value: str, obj):
    glom_comparator = get_glom_comparator(glom_path, cmp_str, cmp_value)
    if glom_comparator(obj):
        obj = assoc(obj, RETAIN, True)
    return obj


def filter_del_lines_from_load_annotations(call):
    return call.get("alt") != "<DEL>"


def filter_bnd_cpx_ctx_lns_from_load_annotations(call):
    return call.get("mutation_type") not in ("BND", "CPX", "CTX")



import re
from cytoolz.curried import assoc, compose
from genomoncology.parse.doctypes import DocType, __CHILD__, __TYPE__
from genomoncology.pipeline.transformers import register, name_mapping


NAME_MAPPING = {
    "gene": "gene_name",
    "alterations": "alterations",
    "CNT__int": "CNT",
    "tissues__mstring": "TISSUES",
    "tissue_cnt__mstring": "TISSUES",
    "tissue_frequency__mstring": "TISSUES_FREQ",
    "tissue_subtype_cnt__mstring": "TISSUES_SUBTYPE",
    "tissue_subtype_frequency__mstring": "TISSUES_SUBTYPE_FREQ",
    "histology_cnt__mstring": "HISTOLOGY",
    "histology_frequency__mstring": "HISTOLOGY_FREQ",
    "histology_subtype_cnt__mstring": "HISTOLOGY_SUBTYPE",
    "histology_subtype_frequency__mstring": "HISTOLOGY_SUBTYPE_FREQ",
}


def get_keys(x, field):
    return list(x.get(field, {}).keys())


def process_frequency_and_counts(x, field):
    freq_results = []
    freq_info = x.get(field)
    for tissue, freq in freq_info.items():
        freq_results.append(f"{tissue}={freq}")
    return freq_results


def remove_start_field(x):
    x.pop("start", "")
    return x


def get_chr(x):
    if x.get("chr"):
        result = re.search(r"(\d{1,2}\b)", x.get("chr", ""))
        return int(result.group())


register(
    input_type=DocType.AGGREGATE,
    output_type=DocType.COSMIC_NON_SNV,
    transformer=compose(
        lambda x: assoc(
            x, "tissues__mstring", get_keys(x, "tissues__mstring")
        ),
        lambda x: assoc(
            x,
            "tissue_cnt__mstring",
            process_frequency_and_counts(x, "tissue_cnt__mstring"),
        ),
        lambda x: assoc(
            x,
            "tissue_frequency__mstring",
            process_frequency_and_counts(x, "tissue_frequency__mstring"),
        ),
        lambda x: assoc(
            x,
            "tissue_subtype_cnt__mstring",
            process_frequency_and_counts(x, "tissue_subtype_cnt__mstring"),
        ),
        lambda x: assoc(
            x,
            "tissue_subtype_frequency__mstring",
            process_frequency_and_counts(
                x, "tissue_subtype_frequency__mstring"
            ),
        ),
        lambda x: assoc(
            x,
            "histology_cnt__mstring",
            process_frequency_and_counts(x, "histology_cnt__mstring"),
        ),
        lambda x: assoc(
            x,
            "histology_frequency__mstring",
            process_frequency_and_counts(x, "histology_frequency__mstring"),
        ),
        lambda x: assoc(
            x,
            "histology_subtype_cnt__mstring",
            process_frequency_and_counts(x, "histology_subtype_cnt__mstring"),
        ),
        lambda x: assoc(
            x,
            "histology_subtype_frequency__mstring",
            process_frequency_and_counts(
                x, "histology_subtype_frequency__mstring"
            ),
        ),
        lambda x: assoc(x, __TYPE__, DocType.COSMIC_NON_SNV.value),
        name_mapping(NAME_MAPPING),
    ),
)

register(
    input_type=DocType.AGGREGATE,
    output_type=DocType.COSMIC_NON_SNV,
    transformer=compose(
        lambda x: assoc(x, __CHILD__, DocType.COSMIC_NON_SNV.value)
    ),
    is_header=True,
)



import tarfile

from click._compat import open_stream
from click.exceptions import FileError

from .const import TGZ_PATH_DELIMITER


class SmartLazyFile(object):
    """
    Built using LazyFile from Click project. Reference:
    https://github.com/pallets/click/blob/master/click/utils.py#L70
    """

    def __init__(
        self, filename, mode="r", encoding=None, errors="strict", atomic=False
    ):
        self.name = filename
        self.mode = mode
        self.encoding = encoding
        self.errors = errors
        self.atomic = atomic
        self.member_path = None

        if filename == "-":
            self._f, self.should_close = open_stream(
                filename, mode, encoding, errors
            )

        elif filename.startswith(f"tgz{TGZ_PATH_DELIMITER}"):
            assert "w" not in mode, "TGZ not supporting write mode."

            (_, file_path, member_path) = filename.split(TGZ_PATH_DELIMITER)
            tgz = tarfile.open(file_path)
            self.name = file_path
            self.member_path = member_path

            # note: decode(): tarball files come out as bytes not strings
            self._f = map(lambda x: x.decode(), tgz.extractfile(member_path))
            self.should_close = False

        else:
            if "r" in mode:
                # Open and close the file in case we're opening it for
                # reading so that we can catch at least some errors in
                # some cases early.
                open(filename, mode).close()
            self._f = None
            self.should_close = True

    def __getattr__(self, name):
        return getattr(self.open(), name)

    def __repr__(self):
        if self._f is not None:
            return repr(self._f)
        return "<unopened file %r %s>" % (self.name, self.mode)

    def open(self):
        """Opens the file if it's not yet open.  This call might fail with
        a :exc:`FileError`.  Not handling this error will produce an error
        that Click shows.
        """
        if self._f is not None:
            return self._f
        try:
            rv, self.should_close = open_stream(
                self.name,
                self.mode,
                self.encoding,
                self.errors,
                atomic=self.atomic,
            )
        except (IOError, OSError) as e:
            raise FileError(self.name, hint=self._get_streerror(e))
        self._f = rv
        return rv

    def _get_streerror(e, default=None):
        """ had to add recursion error handling. In event that attribute doesnt
        exist in object, it causes recursion error."""
        try:
            if hasattr(e, 'strerror'):
                msg = e.strerror
            else:
                if default is not None:
                    msg = default
                else:
                    msg = str(e)
            if isinstance(msg, bytes):
                msg = msg.decode('utf-8', 'replace')
            return msg
        except RecursionError as e:
            msg = f'Recursion error - strerror does not exist {e}'
            return msg

    def close(self):
        """Closes the underlying file, no matter what."""
        if self._f is not None:
            self._f.close()

    def close_intelligently(self):
        """This function only closes the file if it was opened by the lazy
        file wrapper.  For instance this will never close stdin.
        """
        if self.should_close:
            self.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, tb):
        self.close_intelligently()

    def __iter__(self):
        self.open()
        return iter(self._f)



from cytoolz.curried import assoc, compose
from genomoncology.parse.doctypes import DocType, __CHILD__, __TYPE__
from genomoncology.pipeline.transformers import (
    register,
    name_mapping,
)


NAME_MAPPING = {
    "csra": "mnv",
    "AC_mnv__float": "AC_mnv",
    "AF_mnv__float": "AF_mnv"
}


def get_csra(x):
    piped_csra = x['csra'].replace('-', '|')
    return f'chr{piped_csra}|GRCh37'


register(
    input_type=DocType.TSV,
    output_type=DocType.GNOMAD_MNV,
    transformer=compose(
        lambda x: assoc(x, __TYPE__, DocType.GNOMAD_MNV.value),

        lambda x: assoc(x, "csra", get_csra(x)),
        name_mapping(NAME_MAPPING),
    ),
)
register(
    input_type=DocType.TSV,
    output_type=DocType.GNOMAD_MNV,
    transformer=compose(
        lambda x: assoc(x, __CHILD__, DocType.GNOMAD_MNV.value)
    ),
    is_header=True,
)



import asyncio
import logging
import uuid
from multiprocessing import Process, Queue

from bravado_core.exception import SwaggerError
from cytoolz.curried import assoc, curry
from genomoncology import kms
from genomoncology.parse import DocType
from genomoncology.pipeline.transformers import filter_private_keys
from glom import Coalesce, glom

from . import Sink

_logger = logging.getLogger(__name__)

EOL = "EOL"


async def make_call(data_set, data_set_version, build, in_queue,
                    is_custom=False):
    unit = in_queue.get()

    while unit != EOL:
        try:
            _, response = await kms.annotations.load_annotations(
                unit,
                data_set=data_set,
                data_set_version=data_set_version,
                build=build,
                is_custom=is_custom
            )
        except (SwaggerError, OSError, ConnectionError, TimeoutError) as e:
            response = {}
            _logger.exception("failed to load annotation: %s", e)
        for e in response.get("failed_to_load", []):
            _logger.error("failed to load annotation: %s", e)
        for e in response.get("loaded_but_with_errors", []):
            _logger.error("annotation loaded with an error: %s", e)

        unit = in_queue.get()


def background_runner(data_set, data_set_version, build, queue, is_custom):
    asyncio.run(make_call(data_set, data_set_version, build, queue, is_custom))


@curry
class LoadAnnotationSink(Sink):
    def __init__(self, _, state, data_set, data_set_version, num_workers=10,
                 is_custom=False):
        self.data_set = data_set
        self.data_set_version = data_set_version
        self.queue = Queue(maxsize=1000)
        self.workers = []
        self.is_custom = is_custom
        self.num_workers = num_workers

        for _ in range(num_workers):
            worker = Process(
                target=background_runner,
                args=(
                    data_set, data_set_version, state.build, self.queue,
                    self.is_custom),
            )
            worker.start()
            self.workers.append(worker)

    def close(self):
        for _ in range(self.num_workers):
            self.queue.put(EOL)

        for worker in self.workers:
            worker.join()

    def write(self, unit):
        unit = filter(lambda x: not DocType.HEADER.is_a(x), unit)
        unit = list(map(filter_private_keys, unit))

        self.queue.put(unit)
        return unit


UUID = "uuid"


@curry
def set_uuid(data_set, data_set_version, record):
    if UUID not in record:
        uuid_value = glom(record, Coalesce("hgvs_g", "hgvs_c", default=None))

        if uuid_value:
            uuid_value = f"{data_set}|{data_set_version}|{uuid_value}"
            uuid_value = str(uuid.uuid3(uuid.NAMESPACE_DNS, uuid_value))
        else:
            uuid_value = str(uuid.uuid4())

        record = assoc(record, UUID, uuid_value)

    return record



from cytoolz.curried import assoc, compose

from genomoncology.parse.doctypes import DocType, __TYPE__, __CHILD__
from genomoncology.pipeline.transformers.functions import filter_keys, rename
from genomoncology.pipeline.transformers.registry import register


register(
    input_type=DocType.ANNOTATED_CALL,
    output_type=DocType.MARKER,
    transformer=compose(
        assoc(key=__TYPE__, value=DocType.MARKER.value),
        filter_keys(
            {
                "alt_depth",
                "is_het",
                "is_phased",
                "is_somatic",
                "mutation_hash",
                "quality",
                "ref_depth",
                "total_depth",
                "vaf",
                "vaf_alt",
            }
        ),
        rename("annotations.hgvs_g", "mutation_hash"),
    ),
)

register(
    input_type=DocType.ANNOTATED_CALL,
    output_type=DocType.MARKER,
    transformer=assoc(key=__CHILD__, value=DocType.MARKER.value),
    is_header=True,
)
